// fibo.cpp: parallel Fibonacci benchmark
//
// modifications:
//
// 09/30/16 RD	calculate by default fibo(45), 32 units (release v030)
// 09/30/16 RD	use sc_fifo channels for communication
// 06/03/16 RD	added reference function Fibonacci for result validation
// 06/03/16 RD	patches to work around current RISC limitations
// 05/31/16 RD	initial SystemC benchmark (derived from SpecC example)

#include <stdio.h>
#include <stdlib.h>
#include "systemc.h"

// number of parallel instances (powers of 2 in range 1 to 1024)
#ifndef PAR
#define PAR 32
#endif

// type of Fibonacci numbers
#define num_t unsigned long long

SC_MODULE(Fibo1)
{
    sc_fifo_in<num_t>	portN;
    sc_fifo_out<num_t>	portFibN;

    num_t fibo(num_t n)
    {
	if (n <= 1)
	    return n;
	else
	    return fibo(n-1) + fibo(n-2); 
    }

    void main(void)
    {
	num_t N, FibN;

	portN.read(N);
#ifdef DEBUG
	fprintf(stderr, "%s(%lld)\n", name(), N);
#endif
	FibN = fibo(N);
	portFibN.write(FibN);
    }

    SC_CTOR(Fibo1)
    {
	SC_THREAD(main);
    }
};

#define FIBO_PAR(NAME, COMPONENT)		\
SC_MODULE(NAME)					\
{						\
    sc_fifo_in<num_t>	portN;			\
    sc_fifo_out<num_t>	portFibN;		\
						\
    sc_fifo<num_t>	n1, n2, f1, f2;		\
    COMPONENT		fib1,			\
			fib2;			\
						\
    void main(void)				\
    {						\
	num_t N, FibN;				\
						\
	portN.read(N);				\
	if (N <= 1)				\
	{					\
	    FibN = N;				\
	}					\
	else					\
	{					\
	    n1.write(N-1);			\
	    n2.write(N-2);			\
	    FibN = f1.read() + f2.read();	\
	}					\
	portFibN.write(FibN);			\
    }						\
						\
    SC_CTOR(NAME):				\
	fib1("fib1"),				\
	fib2("fib2")				\
    {						\
	SC_THREAD(main);			\
    }						\
	void before_end_of_elaboration() \
	{\
	fib1.portN.bind(n1);			\
	fib2.portN.bind(n2);			\
	fib1.portFibN.bind(f1);			\
	fib2.portFibN.bind(f2);			\
  } \
};

#if PAR >= 2
FIBO_PAR(Fibo2, Fibo1)
#define TOP Fibo2
#endif

#if PAR >= 4
FIBO_PAR(Fibo4, Fibo2)
#undef TOP
#define TOP Fibo4
#endif

#if PAR >= 8
FIBO_PAR(Fibo8, Fibo4)
#undef TOP
#define TOP Fibo8
#endif

#if PAR >= 16
FIBO_PAR(Fibo16, Fibo8)
#undef TOP
#define TOP Fibo16
#endif

#if PAR >= 32
FIBO_PAR(Fibo32, Fibo16)
#undef TOP
#define TOP Fibo32
#endif

#if PAR >= 64
FIBO_PAR(Fibo64, Fibo32)
#undef TOP
#define TOP Fibo64
#endif

#if PAR >= 128
FIBO_PAR(Fibo128, Fibo64)
#undef TOP
#define TOP Fibo128
#endif

#if PAR >= 256
FIBO_PAR(Fibo256, Fibo128)
#undef TOP
#define TOP Fibo256
#endif

#if PAR >= 512
FIBO_PAR(Fibo512, Fibo256)
#undef TOP
#define TOP Fibo512
#endif

#if PAR >= 1024
FIBO_PAR(Fibo1024, Fibo512)
#undef TOP
#define TOP Fibo1024
#endif

SC_MODULE(Main)
{
    num_t		N, FibN;
    sc_fifo<num_t>	n, fib;
    TOP			fibo;

    void main(void)
    {
	n.write(N);
	FibN = fib.read();
    }

    SC_CTOR(Main):
	fibo("fibo")
    {
	SC_THREAD(main);
    }

    void before_end_of_elaboration()
    {
	fibo.portN.bind(n);
	fibo.portFibN.bind(fib);
    }
};

num_t Fibonacci(num_t n) // fast reference function (11/09/11, RD)
{
   const num_t	MEM_SIZE = 100;
   static num_t	fib[MEM_SIZE] = {0,1}; /* memory */

   if (n <= 1)	/* base case */
      { return n;
       } /* fi */
   else		/* previouly calculated results */
      { if (n < MEM_SIZE && fib[n])
	   { return fib[n];
	    } /* fi */
	else	/* recursion step */
	   { num_t f;
	     f = Fibonacci(n-1) + Fibonacci(n-2);
	     if (n < MEM_SIZE)
		{ fib[n] = f;	/* remember this */
		 } /* fi */
	     return f;
	    } /* esle */
       } /* esle */
} /* end of Fibonacci */

int sc_main(int argc, char* argv[])
{
    num_t	n, fib;
    Main	main("main");

    if (argc == 2)
    {
	n = atoi(argv[1]);
    }
    else
    {
	n = 45;
    }
    printf("Fibonacci[%d,%d] starting...\n", PAR, (int)n);
    main.N = n;
    sc_start();
    fib = main.FibN;
    printf("Fibonacci(%lld) = %lld\n", n, fib);
    if (fib == Fibonacci(n))
    {
	printf("Correct. Done!\n");
	return(0);
    }
    else
    {
	printf("Incorrect, Fibonacci(%lld) = %lld. Failed!\n",
		n, Fibonacci(n));
	return(10);
    }
}

// EOF fibo.cpp
