/* risc.cpp: RISC Compiler (c) 2018 CECS (TS, GL, RD) */

#include <iostream>
#include <ctime>

#include "risc.h"


// RISC debugging features based on hints by TS (06/03/18, RD)

#ifndef NDEBUG  // available only in DEBUG mode (in gdb/ddd, e.g. when RISC has crashed)

void DebugAST(            // .dot-file generator to inspect any SgNode's AS
  const SgNode *AnyNode)  // -> GDB: call DebugAST(Pointer_to_any_SgNode)
{                         // -> csh: xdot.py AST.dot
  AstDOTGeneration ast;
  ast.generate((SgNode*)AnyNode, "AST");
  printf("DEBUG: file \"AST.dot\" generated for: %s\n",
           AnyNode->unparseToString().c_str());
}

#endif  // !NDEBUG


#define HELP_PAGE                                                                   \
"risc: " RISC_LIBRARY_NAME " Compiler, V" RISC_VERSION_NUMBER "\n"                  \
"      " RISC_COPYRIGHT "\n"                                                        \
"\n"                                                                                \
"Usage: risc [<option>...] <design> [<option>...]\n"                                \
"\n"                                                                                \
"Options:               Description:                    Default:\n"                 \
"-h|--help              print this usage info, quit     none\n"                     \
"-v|--verbose           increment verbosity             silent\n"                   \
"-vv                    increment verbosity by 2        silent\n"                   \
"-vvv                   increment verbosity by 3        silent\n"                   \
"-w|--warnings          increment warning level         no warnings\n"              \
"-ww                    increment warning level by 2    no warnings\n"              \
"-www                   increment warning level by 3    no warnings\n"              \
"-g                     produce debugging information   no debugging infos\n"       \
"-O|-O<level>           enable optimizations            no optimization\n"          \
"-I<dir>                add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-L<dir>                add dir to library path         $SYSTEMC_OOP_HOME/lib\n"    \
"-l<lib>                add lib to linker libraries     -lsystemc\n"                \
"-c                     compile only, do not link       compile and link\n"         \
"-o <filename>          name output file as filename    a.out\n"                    \
"-psg                   turn on PSG generation mode     none\n"                     \
"-psg_input             set input psg file name         none\n"                     \
"-psg_output            set output psg file name        <design>.psg\n"             \
"-risc:dump             output SG, conflict tables      none\n"                     \
"-risc:icpc             use Intel compiler backend      g++\n"                      \
"-risc:mic              compile for MIC architecture    none\n"                     \
"-risc:elab <filename>  import elaboration result       none\n"                     \
"<rose:option>          passed to ROSE compiler         none\n"                     \
"<g++_option>           passed to GNU C++ compiler      none\n"                     \
"\n"
// TODO: add "-E        preprocess only, do not compile compile and link\n" \
// TODO: how? replace "-E" option with "-rose:unparse_includes -rose:unparse_line_directives" (and if possible -rose:exit_after_preprocessing or similar)

class RISCParseStatus: public risc::tools::ParseStatus
{
  public:
    bool dump;
    bool psg;
    
    RISCParseStatus(): ParseStatus(), dump(false), psg(false) {}
    RISCParseStatus& operator=(const RISCParseStatus& copy)
    {
      verbosityLevel = copy.verbosityLevel;
      warningLevel = copy.warningLevel;
      compilerName = copy.compilerName;
      mic = copy.mic;
      dump = copy.dump;
      psg = copy.psg;

      return *this;
    }
};

void
RISCParseCommandlineOptions(int &argc, char** &argv, RISCParseStatus* status)
{
  if (argc < 2)
  {
    std::cout << "risc: no input files\n";
    exit(0);
  }

  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;
  bool next_arg_is_input_psg_file_name = false;
  bool next_arg_is_output_psg_file_name = false;
  
  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    if(next_arg_is_input_psg_file_name){
      next_arg_is_input_psg_file_name = false;
      
      status->input_psg_file_names.push_back(*it);
      RemovedIter.push_back(it);
    }
    
    if(next_arg_is_output_psg_file_name){
      next_arg_is_output_psg_file_name = false;
      
      status->output_psg_file_name = *it;
      RemovedIter.push_back(it);
    }
    
    // -h, --help
    if (it->compare("-h") == 0 || it->compare("--help") == 0)
    {
      std::cout << HELP_PAGE;
      exit(0);
    }

    // -risc:icpc
    if (it->compare("-risc:icpc") == 0)
    {
      status->compilerName = "icpc";
      RemovedIter.push_back(it);
    }

    // -risc:mic
    if (it->compare("-risc:mic") == 0)
    {
      status->mic = true;
      RemovedIter.push_back(it);
    }

    // -risc:dump
    if (it->compare("-risc:dump") == 0)
    {
      status->dump = true;
      RemovedIter.push_back(it);
    }
    
    if (it->compare("-psg") == 0)
    {
      status->psg = true;
      RemovedIter.push_back(it);
    }
    
    if (it->compare("-psg_input") == 0)
    {
      next_arg_is_input_psg_file_name = true;
      RemovedIter.push_back(it);
    }
    
    if (it->compare("-psg_output") == 0)
    {
      next_arg_is_output_psg_file_name = true;
      RemovedIter.push_back(it);
    }
    
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);

  // Parse library-level command-line options
  risc::tools::ParseCommandlineOptions(argc, argv, status);
}

int main(int argc, char* argv[])
{
  time_t total_begin = time(0);

  RISCParseStatus status;

  RISCParseCommandlineOptions(argc, argv, &status);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC Compiler starting..."
              << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: adding command-line options for frontend..."
              << std::endl;
  }

  std::string LWIncludePath;
  std::string LWLibraryPath;

  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";

  // add -I$SYSTEMC_LW_HOME/include
  risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);

  // add -L$SYSTEMC_LW_HOME/lib-linux64
  risc::tools::AddCommandlineOption(argc, argv, LWLibraryPath);

  // add -lsystemc
  risc::tools::AddCommandlineOption(argc, argv, "-lsystemc");

  // add -mmic
  if (status.mic == true)
  {
    risc::tools::AddCommandlineOption(argc, argv, "-mmic");
  }
  
  int string_width = 70;
  time_t begin;
  time_t end;

  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout.width(string_width);
    std::cout << std::left << "RISC: parsing, building ROSE IR...";
    std::cout.flush();
    begin = time(0);
  }

  risc::Design d(argc,argv,status.psg);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
    end = time(0);
    time_t elapsed_secs = end - begin;
    std::cout.width(4);
    std::cout << std::right << (int) (elapsed_secs / 60);
    std::cout << "min, ";
    std::cout.width(2);
    std::cout << std::right << (int) (elapsed_secs) % 60;
    std::cout << "sec" << std::endl;
  } else if (status.verbosityLevel == RISC_VERBOSE) {
    std::cout << std::endl;
  }

  std::string design_file_name = d.get_design_file_name() + "_";
  // We measure the width of the output
  // It is just for nicer formating
  std::vector<std::string> text(6);
  text.at(0) = "RISC: adding command-line options for frontend...";
  text.at(1) = "RISC: generating \"" + design_file_name + "current_time_advance_table.html\"...";
  text.at(2) = "RISC: generating \"" + design_file_name + "next_time_advance_table.html\"...";
  text.at(3) = "RISC: generating \"" + design_file_name + "combined_data_conflict_table.html\"...";
  text.at(4) = "RISC: generating \"" + design_file_name + "prediction_event_notification_table.html\"...";
  text.at(5) = "RISC: generating \"" + design_file_name + "prediction_time_advance_table.html\"...";

  for(std::vector<std::string>::iterator iter = text.begin(); iter != text.end(); iter++) {
    if(iter->size() > string_width) {
      string_width = iter->size();
    }
  }

  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "RISC: analyzing, building RISC SG..."
              << std::endl;
  }
  risc::sg::SegmentGraph::FunctionCallBoundarySet function_call_boundary(
      risc::tools::SystemCDefinitions::get_sc_wait_funcs());
  risc::sg::SegmentGraph sg(function_call_boundary);

  if (status.verbosityLevel >= RISC_VERY_VERBOSE)
  {
    std::cout << "RISC: building segment graph..."
              << std::endl;
    begin = time(0);
    std::cout.flush();
  }

  typedef std::pair<int, risc::ThreadVectorIter> IdThreadPair;
  std::list<IdThreadPair> starting_ids;
  for (risc::ModuleVectorIter
       iter_module  = d.module_definitions_.begin();
       iter_module != d.module_definitions_.end();
       iter_module++) {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "  Module: " <<  (*iter_module)->get_name() << std::endl;
    }

    if (!PSG_MODE && (*iter_module)->instance_counter_ == 0) {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "          Skipping segment graph generation, no instances." << std::endl;
      }

      continue;
    }

    if (risc::get_design()->perform_static_analysis()) {

      if ((*iter_module)->has_module_pointers()
          || (*iter_module)->has_channel_pointers()) {

        std::cout << "    Error: This module has a sub module/channel which is an array or pointer." << std::endl;
        std::cout << "           This static analysis cannot handle this situation. A hybrid analysis is required." << std::endl;
        exit(1);
      }
    }

    for (risc::ThreadVectorIter
         iter_thread = (*iter_module)->threads_.begin();
         iter_thread != (*iter_module)->threads_.end();
         iter_thread++) {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "    " << (*iter_thread)->get_name() << std::endl;
      }
      if(!PSG_MODE) sg.add_function(*iter_thread, false, true, "");
      else sg.add_function(*iter_thread, false, true, (*iter_module)->get_qualified_type_name() );
      //starting_ids.push_back(IdThreadPair((*iter_thread)->starting_id_, iter_thread));
    }
  }
  
  if(PSG_MODE)  { 
  //add all functions that are defined in the source file to sg
    for(std::vector<risc::Thread*>::iterator iter = d.loaded_threads.begin();
      iter != d.loaded_threads.end();
      iter ++ )
    {
      sg.add_function(*iter,false,false);
    }	 
  }
  
  std::vector<std::string> input_psg_file_prefix;
  if(!PSG_MODE){
    if((status.input_psg_file_names).size()!=0){
      for(std::vector<std::string>::iterator iter = (status.input_psg_file_names).begin();
        iter != (status.input_psg_file_names).end();
        iter ++)
      {
        input_psg_file_prefix.push_back((*iter).substr(0,(*iter).length()-4));
        sg.read_psg_from_file(*iter);
        if (status.verbosityLevel >= RISC_VERY_VERBOSE)
        {
          std::cout << "    " << "PSG loaded from " << (*iter).substr(0,(*iter).length()-4) << std::endl;
        }
      }
    }
    sg.find_module_segments(d.module_definitions_);
    sg.replace_function_call_vertex_with_its_psg();
    sg.remove_dangling_segments();
    
    for (risc::ModuleVectorIter
      iter_module  = d.module_definitions_.begin();
      iter_module != d.module_definitions_.end();
      iter_module++) 
    {
      for (risc::ThreadVectorIter
           iter_thread = (*iter_module)->threads_.begin();
           iter_thread != (*iter_module)->threads_.end();
           iter_thread++) {
        
        starting_ids.push_back(IdThreadPair((*iter_thread)->starting_id_, iter_thread));
      }
    }    
  }
  
  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
    end = time(0);
    std::cout.width(string_width);
    std::cout << std::left << "RISC: segment graph complete...";

    time_t elapsed_secs = end - begin;
    std::cout.width(4);
    std::cout << std::right << (int) (elapsed_secs / 60);
    std::cout << "min, ";
    std::cout.width(2);
    std::cout << std::right << (int) (elapsed_secs) % 60;
    std::cout << "sec" << std::endl;
  }

  sg.clean_graph();

  if(!sg.is_succesfully_created_) {
    std::cout << "Error: The segment graph could not be created." << std::endl;
    exit(1);
  }

  // generating consecutive numbers
  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;

  if(PSG_MODE){
    
    //consequtive segment id
    int prev_seg_id = 0;
    int continuous_seg_id = 0;
    for(vp = boost::vertices(sg.graph_);
        vp.first != vp.second;
        ++vp.first)
    {
      if(sg.graph_[*vp.first].id_<0) continue;
      int id = sg.graph_[*vp.first].id_;
      if( id == prev_seg_id)
        sg.graph_[*vp.first].id_ = continuous_seg_id;
      else{
        prev_seg_id = sg.graph_[*vp.first].id_;
        continuous_seg_id++;
        sg.graph_[*vp.first].id_ = continuous_seg_id;
      }
      
      //In risc, analyze segment is only called during conflict analysis
      //it is used to put all variables into read or write lisrt 9:41 2018/1/19
      sg.graph_[*vp.first].analyze_segment(); //11:46 2018/1/18
    }
    
    //consequtive node id
    int prev_node_id = 0;
    int continuous_node_id = 0;
    for(vp = boost::vertices(sg.graph_);
        vp.first != vp.second;
        ++vp.first) {
      if(sg.graph_[*vp.first].node_id_<0) continue;
      int id = sg.graph_[*vp.first].node_id_;
      if( id == prev_node_id)
        sg.graph_[*vp.first].id_ = continuous_node_id;
      else{
        prev_seg_id = sg.graph_[*vp.first].node_id_;
        continuous_node_id++;
        sg.graph_[*vp.first].node_id_ = continuous_node_id;
      }
    }
  }else{
    for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second;
        ++vp.first, i++) {

      bool found = false;
      std::list<IdThreadPair>::iterator iter;
      for(iter  = starting_ids.begin();
          iter != starting_ids.end();
          iter++) {

        if((*iter).first == sg.graph_[*vp.first].id_) {
          
          found = true;
          break;
        }
      }

      sg.graph_[*vp.first].id_ = i;
      if(found)  {
        IdThreadPair temp = *iter;
        risc::ThreadVectorIter temp_iter= temp.second;
        (*temp_iter)->starting_id_ = i;
        starting_ids.erase(iter);
      }
    }
  }
  
  int num_psgs = (status.input_psg_file_names).size()+1; //including the main one, so plus one
  std::vector< std::vector<int> > psg_seg_id_array(num_psgs);
  if(!PSG_MODE && num_psgs>1)
  {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "    " << "Preparing the instrumentation id array" << std::endl;
    }
    std::vector<int> psg_seg_size(num_psgs);
    for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second;
      ++vp.first) {
      int tmp_psg_id = sg.graph_[*vp.first].psg_id;
      int tmp_id = sg.graph_[*vp.first].id_;
      int tmp_re_id = sg.graph_[*vp.first].relative_seg_id;
      
      
      
      if(psg_seg_size[tmp_psg_id] < tmp_re_id)
        psg_seg_size[tmp_psg_id] = tmp_re_id;
    }
    
    for(int i = 1;i<num_psgs;i++){
      std::vector<int> v(psg_seg_size[i]+1);
      psg_seg_id_array[i] = v;
    }
    
    for(vp = boost::vertices(sg.graph_);
      vp.first != vp.second;
      ++vp.first) {
      
      int tmp_psg_id = sg.graph_[*vp.first].psg_id;
      int tmp_id = sg.graph_[*vp.first].id_;
      int tmp_re_id = sg.graph_[*vp.first].relative_seg_id;
      //the array will be used in instrumentation later
      psg_seg_id_array[tmp_psg_id][tmp_re_id] = tmp_id;
    }
  }

  if (status.verbosityLevel >= RISC_VERBOSE)
  {
    std::cout << "RISC: analyzing, building conflict tables..."
              << std::endl;
  }

  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
    begin = time(0);
  }
  // Time advance tables
  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {

    std::cout.width(string_width);
    std::cout << std::left << "RISC: building time advance table";
    std::cout.flush();
  }
  risc::sg::TimeAdvanceTable time_advance_table(sg);
  if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
    end = time(0);
    time_t elapsed_secs = end - begin;
    std::cout.width(4);
    std::cout << std::right << (int) (elapsed_secs / 60);
    std::cout << "min, ";
    std::cout.width(2);
    std::cout << std::right << (int) (elapsed_secs) % 60;
    std::cout << "sec" << std::endl;
  }
  if (status.dump == true)
  {
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: generating \"" << design_file_name << "current_time_advance_table.html\"..."
                << std::endl;
    }
    time_advance_table.print_current_time_advance_table(
                          design_file_name + "current_time_advance_table.html");
    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: generating \"" << design_file_name << "next_time_advance_table.html\"..."
                << std::endl;
    }
    time_advance_table.print_next_time_advance_table(
                          design_file_name + "next_time_advance_table.html");
  }

  if(PSG_MODE) 
  {
    
    sg.get_wait_and_notifying_events();
    
    sg.print_psg_file(
            status.output_psg_file_name==""?d.get_design_file_name()+".psg":status.output_psg_file_name);
            
    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "segment_graph.dot\"..."
                  << std::endl;
        sg.print_graph(d.get_design_file_name() + "_segment_graph.dot");
      }
    }
    
    risc::inst::instrumentor(d, sg, 0, 0,
                           0, 0, 0, 0, &status, d.get_design_file_name(),psg_seg_id_array,input_psg_file_prefix);
    
    risc::tools::RemoveCommandlineOption(d, LWIncludePath);

    // remove -L$SYSTEMC_LW_HOME/lib-linux64
    risc::tools::RemoveCommandlineOption(d, LWLibraryPath);

    if (status.mic == true)
    {
      std::string MICIncludePath;
      std::string MICLibraryPath;

      MICIncludePath = "-I";
      MICIncludePath += getenv(SYSTEMC_MIC_HOME_ENV);
      MICIncludePath += "/include";
      MICLibraryPath = "-L";
      MICLibraryPath += getenv(SYSTEMC_MIC_HOME_ENV);
      MICLibraryPath += "/lib-linux64";

      // add -I$SYSTEMC_MIC_HOME/include
      risc::tools::AddCommandlineOption(d, MICIncludePath);

      // add -L$SYSTEMC_MIC_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, MICLibraryPath);
    } else {
      std::string OOPIncludePath;
      std::string OOPLibraryPath;
      std::string OOPLibraryPathwoL;

      OOPIncludePath = "-I";
      OOPIncludePath += getenv(SYSTEMC_OOP_HOME_ENV);
      OOPIncludePath += "/include";
      OOPLibraryPath = "-L";
      OOPLibraryPath += getenv(SYSTEMC_OOP_HOME_ENV);
      OOPLibraryPath += "/lib-linux64";
      OOPLibraryPathwoL = OOPLibraryPath;
      OOPLibraryPathwoL.erase(0, 2);

      // add -I$SYSTEMC_OOP_HOME/include
      risc::tools::AddCommandlineOption(d, OOPIncludePath);

      // add -L$SYSTEMC_OOP_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, OOPLibraryPath);

      // add -Xlinker -R -Xlinker $SYSTEMC_OOP_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, "-Xlinker");
      risc::tools::AddCommandlineOption(d, "-R");
      risc::tools::AddCommandlineOption(d, "-Xlinker");
      risc::tools::AddCommandlineOption(d, OOPLibraryPathwoL);
    }
    risc::inst::risc_unparse(d, &status, "RISC", NULL, NULL);

/*** THIS PATCHING AROUND ROSE LIMITATIONS IS NOT NEEDED ANY MORE
     WITH THE NEW ROSE LIBRARY VERSION 0.9.10.X (RD, 05/30/18)
    risc::inst::risc_patch(d, &status,
        "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
        "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
        "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' " RISC_UNPARSE_FILE_PREFIX "%s",
        "RISC");
***/
    risc::inst::risc_compile(d, &status, "RISC");
    return 0;
  }
  else{
    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: building event conflict table";
      std::cout.flush();
      begin = time(0);
    }

    // Event conflict table
    risc::sg::EventConflictTable
      event_conflict_table(sg, d.path_instance_mapper_, true);

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    }
    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "event_conflict_table.html\"..."
                  << std::endl;
      }
      event_conflict_table.print_conflict_table(design_file_name + "event_conflict_table.html");

      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "conflict_lookup_table.html\"..."
                  << std::endl;
      }
      event_conflict_table.print_lookup_table(design_file_name + "conflict_lookup_table.html");
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: building data conflict table";
      std::cout.flush();
      begin = time(0);
    }

    // Data conflict table
    risc::sg::DataConflictTable data_conflict_table(sg, d.path_instance_mapper_, true);

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    }

    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "data_conflict_table.html\"..."
                  << std::endl;
      }
      data_conflict_table.print_conflict_table(design_file_name + "data_conflict_table.html");

    }

    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "segment_graph.dot\"..."
                  << std::endl;
        sg.print_graph(d.get_design_file_name() + "_segment_graph.dot");
        sg.print_graph_read_write_access(d.get_design_file_name() + "_rw_segment_graph.dot");
      }
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: building combined data conflict table";
      std::cout.flush();
      begin = time(0);
    }

    // Combined data conflict table
    risc::sg::CombinedDataConflictTable cdct(data_conflict_table, d.path_instance_mapper_);

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    }

    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "combined_data_conflict_table.html\"..."
                  << std::endl;
      }
      cdct.print_conflict_table(design_file_name + "combined_data_conflict_table.html");
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: building prediction event notification table";
      std::cout.flush();
      begin = time(0);
    }

    // Prediction event notification table
    risc::sg::PredictionEventNotificationTable pent(event_conflict_table, d.path_instance_mapper_);

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    }
    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "prediction_event_notification_table.html\"..."
                  << std::endl;
      }
      pent.print_notification_table(design_file_name + "prediction_event_notification_table.html");
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: building prediction time advance table";
      std::cout.flush();
      begin = time(0);
    }

    // Prediction time advance table
    risc::sg::PredictionTimeAdvanceTable prediction_time_advance_table(sg, cdct.prediction_steps_);

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    }
    if (status.dump == true)
    {
      if (status.verbosityLevel >= RISC_VERY_VERBOSE)
      {
        std::cout << "RISC: generating \"" << design_file_name << "prediction_time_advance_table.html\"..."
                  << std::endl;
      }
      prediction_time_advance_table.print(design_file_name + "prediction_time_advance_table.html");
    }

    if (status.verbosityLevel >= RISC_VERBOSE)
    {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: instrumenting, generating parallel model...";
      std::cout.flush();
      begin = time(0);
    }
    risc::inst::instrumentor(d, sg, &data_conflict_table, &event_conflict_table,
                             &time_advance_table, &cdct, &prediction_time_advance_table, &pent, &status, d.get_design_file_name(),psg_seg_id_array,
                             input_psg_file_prefix);
    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;
    } else if (status.verbosityLevel == RISC_VERBOSE) {
      std::cout << std::endl;
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: modifying command-line options for backend..."
                << std::endl;
    }

    // Change the command line argument list

    // remove -I$SYSTEMC_LW_HOME/include
    risc::tools::RemoveCommandlineOption(d, LWIncludePath);

    // remove -L$SYSTEMC_LW_HOME/lib-linux64
    risc::tools::RemoveCommandlineOption(d, LWLibraryPath);

    if (status.mic == true)
    {
      std::string MICIncludePath;
      std::string MICLibraryPath;

      MICIncludePath = "-I";
      MICIncludePath += getenv(SYSTEMC_MIC_HOME_ENV);
      MICIncludePath += "/include";
      MICLibraryPath = "-L";
      MICLibraryPath += getenv(SYSTEMC_MIC_HOME_ENV);
      MICLibraryPath += "/lib-linux64";

      // add -I$SYSTEMC_MIC_HOME/include
      risc::tools::AddCommandlineOption(d, MICIncludePath);

      // add -L$SYSTEMC_MIC_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, MICLibraryPath);
    } else {
      std::string OOPIncludePath;
      std::string OOPLibraryPath;
      std::string OOPLibraryPathwoL;

      OOPIncludePath = "-I";
      OOPIncludePath += getenv(SYSTEMC_OOP_HOME_ENV);
      OOPIncludePath += "/include";
      OOPLibraryPath = "-L";
      OOPLibraryPath += getenv(SYSTEMC_OOP_HOME_ENV);
      OOPLibraryPath += "/lib-linux64";
      OOPLibraryPathwoL = OOPLibraryPath;
      OOPLibraryPathwoL.erase(0, 2);

      // add -I$SYSTEMC_OOP_HOME/include
      risc::tools::AddCommandlineOption(d, OOPIncludePath);

      // add -L$SYSTEMC_OOP_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, OOPLibraryPath);

      // add -Xlinker -R -Xlinker $SYSTEMC_OOP_HOME/lib-linux64
      risc::tools::AddCommandlineOption(d, "-Xlinker");
      risc::tools::AddCommandlineOption(d, "-R");
      risc::tools::AddCommandlineOption(d, "-Xlinker");
      risc::tools::AddCommandlineOption(d, OOPLibraryPathwoL);
    }

    // risc backend

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: unparsing source files...";
      std::cout.flush();
      begin = time(0);
    }
    risc::inst::risc_unparse(d, &status, "RISC", NULL, NULL);
    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout.flush();
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout.flush();
      std::cout << "sec" << std::endl;
    }

/*** THIS PATCHING AROUND ROSE LIMITATIONS IS NOT NEEDED ANY MORE
     WITH THE NEW ROSE LIBRARY VERSION 0.9.10.X (RD, 05/30/18)

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: patching ROSE limitations...";
      std::cout.flush();
      begin = time(0);
    }
    risc::inst::risc_patch(d, &status,
        "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
        "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
        "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,"
        "(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process "
        "(\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, "
        "\\5, \\6, \\7)/\' "RISC_UNPARSE_FILE_PREFIX"%s",
        "RISC");
    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout.flush();
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout.flush();
      std::cout << "sec" << std::endl;
    }
***/

    if (status.verbosityLevel >= RISC_VERBOSE)
    {
      std::cout.width(string_width);
      std::cout << std::left << "RISC: compiling executable...";
      std::cout.flush();
      begin = time(0);
    }

    risc::inst::risc_compile(d, &status, "RISC");

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {
      end = time(0);
      time_t elapsed_secs = end - begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout.flush();
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout.flush();
      std::cout << "sec" << std::endl;
    } else if (status.verbosityLevel == RISC_VERBOSE) {
      std::cout << std::endl;
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE)
    {
      std::cout << "RISC: additional information" << std::endl;
    }

    if (status.verbosityLevel >= RISC_VERY_VERBOSE) {

      time_t total_end = time(0);
      std::cout.width(string_width);
      std::cout << std::left << "      total time";

      time_t elapsed_secs = total_end - total_begin;
      std::cout.width(4);
      std::cout << std::right << (int) (elapsed_secs / 60);
      std::cout << "min, ";
      std::cout.width(2);
      std::cout << std::right << (int) (elapsed_secs) % 60;
      std::cout << "sec" << std::endl;

      std::cout << "      " << boost::num_vertices(sg.graph_)  << " segments" << std::endl;
      std::cout << "      " << boost::num_edges(sg.graph_)  << " edges" << std::endl;
      std::cout << "      " << cdct.prediction_steps_ << " prediction steps" << std::endl;

      long long conflicts = 0;
      int size = data_conflict_table.size_of_conflict_table_;
      for(int i = 0; i < size; i++) {

        for(int j = 0; j < size; j++) {

          if(data_conflict_table.conflict_table_[i][j]) {
            conflicts++;
          }
        }
      }

      std::cout << "      Data conflict table has " << ((long)size * (long)size) << " entries." << std::endl;
      std::cout << "        " << conflicts <<  " conflicts" << std::endl;
      std::cout << "        " << ((long)size * (long)size) - conflicts <<  " conflict free" << std::endl;

      conflicts = 0;
      for(int i = 0; i < size; i++) {

        for(int j = 0; j < size; j++) {

          if(event_conflict_table.conflict_table_[i][j]) {
            conflicts++;
          }
        }
      }

      std::cout << "      Event conflict table has " << ((long)size * (long)size) << " entries." << std::endl;
      std::cout << "        " << conflicts <<  " conflicts" << std::endl;
      std::cout << "        " << ((long)size * (long)size) - conflicts <<  " conflict free" << std::endl;

      std::cout << "RISC Compiler done."
                << std::endl;
    }
  }
  return 0;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
