// ----------------------------------------------------------------------------
// visual.cpp
// ----------------------------------------------------------------------------
//
// Author: Daniel Mendoza
//
// 
// ----------------------------------------------------------------------------


#include <gtk/gtk.h>
#include <cairo.h>
#include <stdlib.h>
#include <time.h>
#include <sstream>

#include "risc.h"

#define D_SPACE 20 //needed for module sizing
#define MINW 40
#define MINH 50

#define TEXT_CHAR 8

#define DOT_WIDTH 8

#define TRANS_VAR 0.2

#define PRIM_CHANNEL_DISTANCE 20
#define PORT_TEXT_XOFFSET 5
#define PORT_TEXT_YOFFSET 10
#define COLOR_PALET_FORMATTER 13

static double global_unscaled_top_width = 0; //needed for pdf export
static double global_unscaled_top_height = 0; //needed for pdf export

int format_color_counter(int counter, int max_size) {
	while(counter >= max_size) {
		counter -= max_size;
	}

	return counter;
}

/**
 * \struct	Argument Formatter
 * \brief	This struct handles all commandline parsing
 *
 */
struct ArgumentFormatter {
	int lowest_level;

	bool spec_printing_mod;
	std::string top_mod;

	bool black_white_only;

	double scale_factor;

	bool draw_ports;

	void delete_argv_element(int &cur_index, int &argc, char**argv) {
		for(int i = cur_index; i < argc - 1; i++) {
			argv[i] = argv[i + 1];
		}
		argc = argc - 1;
		cur_index = cur_index - 1;
	}

	ArgumentFormatter(int &argc, char** argv) {
		black_white_only = false;
	
		spec_printing_mod = false;	
		top_mod = "";
		
		lowest_level = 0;
		
		scale_factor = 1;

		int delete_counter = 0;
	
		draw_ports = true;
	
		for(int i = 0; i < argc; i++) {
			std::string str(argv[i]);
			
			if(str == "-bw") {
				black_white_only = true;
				delete_argv_element(i,argc,argv);
			}
			else if(str == "-tm") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
					top_mod = std::string(argv[i]);
					delete_argv_element(i,argc,argv);

				}
			}
			else if(str == "-ll") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
				std::string str(argv[i]);
				std::stringstream intConverter;
				intConverter << str;
				intConverter >> lowest_level;
				delete_argv_element(i,argc,argv);
				}
			}
			else if(str == "-s") {
				delete_argv_element(i,argc,argv);
				if(++i < argc) {
				std::string str(argv[i]);
				std::stringstream doubleConverter;
				doubleConverter << str;
				doubleConverter >> scale_factor;
				delete_argv_element(i,argc,argv);

				}
			}
			else if(str == "-h") {
				print_help_menu();
			}
			else if(str == "-np") {
				delete_argv_element(i,argc,argv);
				draw_ports = false;
			}
		}
	}

	void toggle_ports(){
		draw_ports = !draw_ports;
	}

	bool is_drawing_ports() {
		return draw_ports;
	}

	void print_help_menu(){
		std::cout << "Command line options for visualization tool:" << std::endl;
		std::cout << "-h : Lists command line options and their usages." << std::endl;
		std::cout << "-bw : Modules are drawn without color." << std::endl;
		std::cout << "-tm [module name] : Only draw [module name]." << std::endl;
		std::cout << "-ll [integer] : Draw only a certain depth in the hierarchy given by [integer}." << std::endl;
		std::cout << "-np : Draws module hierarchy without ports or channels." << std::endl;
		std::cout << "-s [floating point number] : Scale the drawing size by [floating point number]. If [floating point number] = 0.5, then the size of the drawing is scaled by 50 percent." << std::endl;
	}

	void increment_depth() {
		if(lowest_level != 0) {
		lowest_level++;
		}
	}

	void decrement_depth() {
		if(lowest_level <= 1) return;
		else {
			lowest_level--;
		}
	}

	void toggle_depth() {
		if(lowest_level != 0) {
			lowest_level = 0;
		}
		else {
			lowest_level = 1;
		}
	}

	int get_depth_var() {
		return lowest_level;
	}

	void increment_scale_factor_by5percent() {
		scale_factor += 0.05;
	}

	void decrement_scale_factor_by5percent() {
		scale_factor -= 0.05;
	}

	void set_scale_factor(double factor) {
		scale_factor = factor;
	}

	double get_scale_factor() {
		return scale_factor;
	}

	bool is_bw_only() {
		return black_white_only;
	}

	void toggle_bw() {
		if(black_white_only) {
			black_white_only = false;
		}
		else {
			black_white_only = true;
		}
	}

	void unset_print_tag(std::string cur_mod) {
		if(spec_printing_mod && cur_mod == top_mod) {
			spec_printing_mod = false;
		}
	}
	
	bool get_set_print_tag_true(std::string cur_mod) {
		if(top_mod != "") {
			if(spec_printing_mod) {
				return true;
			}
			else if(cur_mod == top_mod) {
				spec_printing_mod = true;
				return spec_printing_mod;
			}
			else {
				return false;
			}
		}
		else {
			return true;;
		}
	}

	bool is_this_lowest_level(int cur_level) {
		if(lowest_level <= 0) {
			return false;
		}
		else {
			if(cur_level == lowest_level) {
				return true;
			}
			else {
				return false;
			}
		}
	}

};

/**
 *  \struct	Colorer
 *  \brief	This struct handles the coloring of modules based on their class type
 *
 */
struct Colorer {

	std::vector<risc::Class*> mods;

	int get_color(risc::Instance* mod_param) {
		int i;
		risc::Class* uncolored_mod = mod_param->get_definition();
		for(i = 0; i < mods.size(); i++) {
			if(mods.at(i) == uncolored_mod) {
				return format_color_counter(i,COLOR_PALET_FORMATTER);
			}
		}
		mods.push_back(uncolored_mod);
		return format_color_counter(i,COLOR_PALET_FORMATTER);
		
	}
};

/**
 *  \struct	Box
 *  \brief	This struct represents a module
 *
 */
struct Box {
	
	Box(int x, int y, int w, int h) {
		x_box_cor = x;
		y_box_cor = y;
		box_width = w;
		box_height = h;
	}
	
	int x_box_cor;
	int y_box_cor;
	int box_width;
	int box_height;
};

/**
 * \struct	RGB
 * \brief	This function represents a pixel with its corresponding red, green and blue
 * 		intensities.
 */
struct RGB {
	RGB(double r, double g, double b) {
	r_val = r;
	g_val = g;	
	b_val = b;
	}

	double r_val;
	double g_val;
	double b_val;

};

struct VisualPort {
	VisualPort(int x, int y, risc::Port* port) {
		xcor = x;
		ycor = y;
		port_ptr = port;
	}

	int ycor;
	int xcor;
	risc::Port* port_ptr;
};

void draw_color_fill(cairo_t* cr,
	int xcur,
	int ycur,
	int width,
	int height,
	double color_r,
	double color_g,
	double color_b) {
	
	cairo_set_source_rgba(cr,color_r,color_g,color_b,TRANS_VAR);
	cairo_rectangle(cr,xcur,ycur,width,height);
	cairo_fill(cr);

}

int find_largest_char_port(risc::Instance *m) {
	int largest_string = 0;
	for(risc::InPortVectorIter inport_iter = m->get_definition()->inports_.begin();
		inport_iter != m->get_definition()->inports_.end();
		inport_iter++) {
		if( ( (*inport_iter)->get_name()).length() > largest_string ) {
			largest_string = ( (*inport_iter)->get_name()).length();
		}
	}
	for(risc::OutPortVectorIter outport_iter = m->get_definition()->outports_.begin();
		outport_iter != m->get_definition()->outports_.end();
		outport_iter++) {
			if( ((*outport_iter)->get_name()).length() > largest_string ) {
				largest_string = ( (*outport_iter)->get_name()).length();
			}
		}
	for(risc::InOutPortVectorIter inoutport_iter = m->get_definition()->inoutports_.begin();
		inoutport_iter != m->get_definition()->inoutports_.end();
		inoutport_iter++) {
		if( ( (*inoutport_iter)->get_name()).length() > largest_string ) {
			largest_string = ( (*inoutport_iter)->get_name()).length();
		}
	}
	
	return largest_string;
}

void draw_dot(cairo_t* cr, int x, int y) {
	
	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_set_line_width(cr,DOT_WIDTH);
	cairo_set_line_cap(cr,CAIRO_LINE_CAP_ROUND);
	cairo_move_to(cr,x,y);
	cairo_line_to(cr,x,y);
	cairo_stroke(cr);
}

void draw_port(cairo_t* cr,int xcur, int ycur,risc::Port* port) {
	
	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	//draw_dot(cr,xcur + max_distance/2,ycur);
	draw_dot(cr,xcur,ycur);	

	//cairo_move_to(cr,xcur + max_distance/2 - ( port->get_name()).length()/2 * TEXT_CHAR,ycur + 10);
	cairo_move_to(cr,xcur + PORT_TEXT_XOFFSET,ycur + PORT_TEXT_YOFFSET);
	cairo_set_font_size(cr,10);
	cairo_show_text(cr,( port->get_name()).c_str());
}

int draw_base_ports_wrapper(cairo_t* cr,int x, int y, risc::Instance *m, std::vector<VisualPort> &VisualPortVector) {
	int port_distance = find_largest_char_port(m) * TEXT_CHAR;
	int total_x_offset = 0;
	for(risc::InPortVectorIter inport_iter = m->get_definition()->inports_.begin();
		inport_iter != m->get_definition()->inports_.end();
		inport_iter++) {
		draw_port(cr,x,y,(*inport_iter));
		
		VisualPortVector.push_back(VisualPort(x,y,(*inport_iter)));

		x += port_distance + PORT_TEXT_XOFFSET;
		total_x_offset += port_distance + PORT_TEXT_XOFFSET;


	}
        for(risc::OutPortVectorIter outport_iter = m->get_definition()->outports_.begin();
                outport_iter != m->get_definition()->outports_.end();
                outport_iter++) {
                draw_port(cr,x,y,(*outport_iter));

		VisualPortVector.push_back(VisualPort(x,y,(*outport_iter)));

                x += port_distance + PORT_TEXT_XOFFSET;
                total_x_offset += port_distance + PORT_TEXT_XOFFSET;

        }
        for(risc::InOutPortVectorIter inoutport_iter = m->get_definition()->inoutports_.begin();
                inoutport_iter != m->get_definition()->inoutports_.end();
                inoutport_iter++) {
                draw_port(cr,x,y,(*inoutport_iter));
                
		VisualPortVector.push_back(VisualPort(x,y,(*inoutport_iter)));

		x += port_distance + PORT_TEXT_XOFFSET;
                total_x_offset += port_distance + PORT_TEXT_XOFFSET;
		
        }
	
	return total_x_offset;
}

void draw_connect_line(cairo_t* cr, int x1, int y1, int x2, int y2) {
	cairo_set_line_width(cr,1);
	cairo_set_source_rgb(cr,0,0,0);
	cairo_move_to(cr,x1,y1);
	cairo_line_to(cr,x1,y2/2);
	cairo_line_to(cr,x2,y2/2);
	cairo_line_to(cr,x2,y2);
	cairo_stroke(cr);
}

void draw_line(cairo_t* cr, int x1, int y1, int x2, int y2) {
	
	static bool collision_manager = false;
	if(cr == NULL) {
		collision_manager = false;
		return;
	}
	else if(cr != NULL 
			&& x1 == 0
			&& x2 == 0
			&& y1 == 0
			&& y2 == 0) {
					collision_manager = true;
					return;
	}


	cairo_set_line_width(cr,1);
	cairo_set_source_rgb(cr,0,0,0);
	cairo_move_to(cr,x1,y1);

	if(x1 == x2 || y1 == y2) {
		cairo_line_to(cr,x2,y2);
		collision_manager = false;
	}
	else {
		int y_distance = y2 - y1;
		double scaler;

		if(collision_manager) {
			scaler = 0.75;
			collision_manager = false;
		}
		else {
			collision_manager = true;
			scaler = 0.5;
		}

		cairo_line_to(cr,x1,y1 + y_distance*scaler);
		
		cairo_move_to(cr,x1,y1 + y_distance*scaler);
		cairo_line_to(cr,x2,y1 + y_distance*scaler);
		
		cairo_move_to(cr,x2,y1 + y_distance*scaler);
		cairo_line_to(cr,x2,y2);

	}
	cairo_stroke(cr);
}

/*
 * \fn		draw_compared_ports
 * \brief	This function draws the ports on the module.
 */
int draw_compared_ports(cairo_t* cr, int x, int y, std::vector<VisualPort> &lower_ports,std::vector<VisualPort> &current_ports, risc::Instance *current_mod, risc::Instance *lower_mod,int &prev_end) {
	int port_distance = find_largest_char_port(current_mod) * TEXT_CHAR;
	int temp_prev_port_xcor = 0;
	int temp_prev_port_name = 0;
	int temp_adder = 0;
	int max_xcor = 0;
	SgVariableSymbol* temp_low_symbol;
	SgVariableSymbol* temp_high_symbol;
	SgVariableSymbol* symbol_comparee;
	bool port_found = false;
	//first iterate through child ports
	for(int i = 0; 
		i < lower_ports.size();
		i++) {
						
			temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
			temp_high_symbol = get_mapped_symbol(temp_low_symbol,
					dynamic_cast<risc::Module*>(current_mod->get_definition()),
					lower_mod->get_ast_node());
		
			for(risc::InPortVectorIter inport_iter = current_mod->get_definition()->inports_.begin();
				inport_iter != current_mod->get_definition()->inports_.end();
				inport_iter++) {

						symbol_comparee = (*inport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
								temp_adder =  temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*inport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*inport_iter)));
							port_found = true;
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*inport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							break;
						}
			}

			if(port_found == true) {
				port_found = false;
				continue;
			}

			for(risc::OutPortVectorIter outport_iter = current_mod->get_definition()->outports_.begin();
				outport_iter != current_mod->get_definition()->outports_.end();
				outport_iter++) {

						symbol_comparee = (*outport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
								temp_adder = temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*outport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*outport_iter)));
							port_found = true;
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*outport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							break;
						}
			}

			if(port_found == true) {
				port_found = false;
				continue;
			}

			for(risc::InOutPortVectorIter inoutport_iter = current_mod->get_definition()->inoutports_.begin();
				inoutport_iter != current_mod->get_definition()->inoutports_.end();
				inoutport_iter++) {

						symbol_comparee = (*inoutport_iter)->get_symbol();
						if(temp_high_symbol == symbol_comparee) {
							if(lower_ports.at(i).xcor - temp_prev_port_xcor < temp_prev_port_name) {
									temp_adder = temp_prev_port_name + PORT_TEXT_XOFFSET;
							}
							else temp_adder = 0;
							if(lower_ports.at(i).xcor + temp_adder < prev_end) {
								temp_adder = prev_end - lower_ports.at(i).xcor;
							}
							draw_port(cr,lower_ports.at(i).xcor + temp_adder,y,(*inoutport_iter));
							draw_line(cr,lower_ports.at(i).xcor + temp_adder,y,lower_ports.at(i).xcor,lower_ports.at(i).ycor);
							current_ports.push_back(VisualPort(lower_ports.at(i).xcor + temp_adder,y,(*inoutport_iter)));
							temp_prev_port_xcor = lower_ports.at(i).xcor + temp_adder;
							temp_prev_port_name = (*inoutport_iter)->get_name().length() * TEXT_CHAR;
							prev_end = temp_prev_port_xcor + temp_prev_port_name;
							if(max_xcor < lower_ports.at(i).xcor + temp_adder + temp_prev_port_name) {
								max_xcor = lower_ports.at(i).xcor + temp_adder + temp_prev_port_name;
							}
							break;
						}
			}
	}
	
	return max_xcor;
}

void draw_box(cairo_t* cr,int x, int y, int width, int height, risc::Instance *m) {
	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_set_line_width(cr,2);
	cairo_rectangle(cr,x,y,width,height);
	cairo_stroke(cr);

	cairo_move_to(cr,x,y + 10);
	cairo_set_font_size(cr,10);
	if(height != MINH) cairo_show_text(cr,(m->get_ast_type_name()+ " " + m->get_name()).c_str());
	else {
		cairo_show_text(cr,(m->get_ast_type_name().c_str()));
		cairo_move_to(cr,x + 5,y + 20);
		cairo_show_text(cr,(m->get_name().c_str()));
	}
}	

bool initialize_color_palet(std::vector<RGB> &color_palet) {
	color_palet.push_back(RGB(1,0,0));
	color_palet.push_back(RGB(0,1,0));
	color_palet.push_back(RGB(0,0,1));
	color_palet.push_back(RGB(0,1,1));
	color_palet.push_back(RGB(1,0,1));
	color_palet.push_back(RGB(1,1,0));
	color_palet.push_back(RGB(1,1,1));
	color_palet.push_back(RGB(0,0.5,1));
	color_palet.push_back(RGB(1,0,0.5));
	color_palet.push_back(RGB(0.5,1,0));
	color_palet.push_back(RGB(0.1,0,0));
	color_palet.push_back(RGB(0,0.1,0));
	color_palet.push_back(RGB(0,0,0.1));

	return true;
}

void color_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box>* bounds_param,
	std::vector<RGB> &color_palet,
	int color_select) {

	int max_xcor = xcur + max_width;
	
	RGB &cur_color = color_palet.at(color_select);

	if(bounds_param == NULL) {
		draw_color_fill(cr,xcur,ycur,max_width,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
		return;
	}

	std::vector<Box> &bounds = (*bounds_param);

	for(std::vector<Box>::iterator iter = bounds.begin();
		iter != bounds.end();
		iter++) {
			
			draw_color_fill(cr,xcur,ycur,iter->x_box_cor - xcur,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			draw_color_fill(cr,iter->x_box_cor,ycur,iter->box_width,iter->y_box_cor - ycur,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			draw_color_fill(cr,iter->x_box_cor,iter->y_box_cor + iter->box_height,iter->box_width,ycur + max_height - (iter->y_box_cor + iter->box_height),cur_color.r_val,cur_color.g_val,cur_color.b_val);
			xcur += iter->box_width + (iter->x_box_cor - xcur);
			
			if(&(*iter) == &bounds.back()) {
				draw_color_fill(cr,xcur,ycur,max_xcor - xcur,max_height,cur_color.r_val,cur_color.g_val,cur_color.b_val);
			}
		
	}
}

void color_level_dependent_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box>* bounds,
	std::vector<RGB> &color_palet,
	int counter) {
	
	counter = format_color_counter(counter,color_palet.size());

	color_box(cr,xcur,ycur,max_width,max_height,bounds,color_palet,counter);

}

void color_module_dependent_box(cairo_t* cr, 
	int xcur,
	int ycur,
	int max_width,
	int max_height,
	std::vector<Box> *bounds,
	std::vector<RGB> &color_palet,
	Colorer &colorer,
	risc::Instance *mod) {
	
	int selector = colorer.get_color(mod);

	color_box(cr,xcur,ycur,max_width,max_height,bounds,color_palet,selector);

}	

int draw_space_channels(risc::Instance *m,int &ycur) {
	int total_offset;
	if(!m->get_definition()->sub_prim_channel_.empty()
		|| !m->get_definition()->sub_hier_channel_.empty()) {
		total_offset = D_SPACE;
	}
	else {
		return 0;
	}

	for(risc::PrimitiveChannelInstanceVectorIter channel_iter = m->get_definition()->sub_prim_channel_.begin();
		channel_iter != m->get_definition()->sub_prim_channel_.end();
		channel_iter++) {
				total_offset += PRIM_CHANNEL_DISTANCE;
	}
	for(risc::HierarchicalChannelInstanceVectorIter channel_iter = m->get_definition()->sub_hier_channel_.begin();
		channel_iter != m->get_definition()->sub_hier_channel_.end();
		channel_iter++) {
				total_offset += PRIM_CHANNEL_DISTANCE;
	}
	ycur += total_offset;
	return total_offset;
}

void draw_prim_channel(cairo_t* cr,int xcur, int ycur,int width,risc::Instance* prim_channel) {
	
	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_move_to(cr,xcur,ycur);
	cairo_set_font_size(cr,10);
	cairo_show_text(cr,( prim_channel->get_name()).c_str());

	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_move_to(cr,xcur + prim_channel->get_name().length() * TEXT_CHAR,ycur);
	cairo_line_to(cr,xcur + width - 2 * D_SPACE,ycur);
	cairo_stroke(cr);
}

void draw_channel_connect(cairo_t* cr, std::vector<VisualPort> &lower_ports, risc::Instance *cur_mod, risc::Instance * lower_mod,int temp_ycur_offset, int y) {
	SgVariableSymbol* temp_low_symbol;
	SgVariableSymbol* temp_high_symbol;
	SgVariableSymbol* symbol_comparee;

	for(risc::HierarchicalChannelInstanceVectorIter channel_iter = cur_mod->get_definition()->sub_hier_channel_.begin();
		channel_iter != cur_mod->get_definition()->sub_hier_channel_.end();
		channel_iter++) {
		symbol_comparee = (*channel_iter)->get_symbol();

		for(int i = 0; i < lower_ports.size(); i++) {

					temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
					temp_high_symbol = get_mapped_symbol(temp_low_symbol,
						dynamic_cast<risc::Module*>(cur_mod->get_definition()),
						lower_mod->get_ast_node());

					if(symbol_comparee == temp_high_symbol) {
						draw_dot(cr,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
						draw_line(cr,lower_ports.at(i).xcor,y + D_SPACE,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
					}
		}
		temp_ycur_offset -= PRIM_CHANNEL_DISTANCE;
	}

	for(risc::PrimitiveChannelInstanceVectorIter channel_iter = cur_mod->get_definition()->sub_prim_channel_.begin();
		channel_iter != cur_mod->get_definition()->sub_prim_channel_.end();
		channel_iter++) {
		symbol_comparee = (*channel_iter)->get_symbol();

		for(int i = 0; i < lower_ports.size(); i++) {

					temp_low_symbol = lower_ports.at(i).port_ptr->get_symbol();
					temp_high_symbol = get_mapped_symbol(temp_low_symbol,
						dynamic_cast<risc::Module*>(cur_mod->get_definition()),
						lower_mod->get_ast_node());

					if(symbol_comparee == temp_high_symbol) {
						draw_dot(cr,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
						draw_line(cr,lower_ports.at(i).xcor,y + D_SPACE,lower_ports.at(i).xcor,y - temp_ycur_offset + D_SPACE);
					}
		}
		temp_ycur_offset -= PRIM_CHANNEL_DISTANCE;
	}
}

/*
 * \fn		draw_channels_wrapper
 * \brief	This function handles the drawing of channels
 */
void draw_channels_wrapper(cairo_t* cr, risc::Instance *m, int x, int y, int mod_width,int temp_ycur_offset, std::vector<VisualPort> &VisualPortVector) {
	if(!m->get_definition()->sub_prim_channel_.empty() 
		|| !m->get_definition()->sub_hier_channel_.empty()) {
		x += D_SPACE;
		y += D_SPACE;
		for(risc::PrimitiveChannelInstanceVectorIter channel_iter = m->get_definition()->sub_prim_channel_.begin();
			channel_iter != m->get_definition()->sub_prim_channel_.end();
			channel_iter++) {
					draw_prim_channel(cr,x,y,mod_width,(*channel_iter));
					y += PRIM_CHANNEL_DISTANCE;
		}

		for(risc::HierarchicalChannelInstanceVectorIter channel_iter = m->get_definition()->sub_hier_channel_.begin();
			channel_iter != m->get_definition()->sub_hier_channel_.end();
			channel_iter++) {
					draw_prim_channel(cr,x,y,mod_width,(*channel_iter));
					y += PRIM_CHANNEL_DISTANCE;
		}
	}
}

/**
 * \fn		draw_recursor
 * \brief	This function is the recursive function that handles the
 * 		control flow of the entire drawing
 */
int draw_recursor(risc::Instance *m, int &param_deepest_level, int &height_tracker, std::vector<VisualPort> &VisualPortVector, cairo_t* cr, ArgumentFormatter &args_format) {

	static int cur_xcor = 0;
	static int cur_ycor = 0;
	static int cur_level = 0;
	static std::vector<RGB> color_palet;
	static Colorer colorer;

	static bool initializer = initialize_color_palet(color_palet);
	
	
	if(args_format.get_set_print_tag_true(m->get_name())) {
		cur_xcor += D_SPACE;
		cur_ycor += D_SPACE;
		cur_level++;

		if(param_deepest_level < cur_level) {
			param_deepest_level = cur_level;
		}
	}
	if(m->get_definition()->sub_modules_.empty()
		&& m->get_definition()->sub_hier_channel_.empty()
		&& m->get_definition()->sub_prim_channel_.empty()
		&& args_format.get_set_print_tag_true(m->get_name())
		|| (args_format.is_this_lowest_level(cur_level)
		&& args_format.get_set_print_tag_true(m->get_name())) ) {
		int base_box_width = 0;
		RGB cur_color = color_palet.at(format_color_counter(cur_level,color_palet.size()));
		
		if(m->get_ast_type_name().length() > m->get_name().length()) {
			base_box_width = m->get_ast_type_name().length() * TEXT_CHAR;
		}
		else {
			base_box_width =  m->get_name().length() * TEXT_CHAR;
		}
	
		if(args_format.is_drawing_ports()) {
		base_box_width += draw_base_ports_wrapper(cr,cur_xcor + base_box_width,cur_ycor,m,VisualPortVector);
		}
		draw_box(cr,cur_xcor,cur_ycor,base_box_width,MINH,m);

		if(!args_format.is_bw_only()) color_module_dependent_box(cr,cur_xcor,cur_ycor,base_box_width,MINH,NULL,color_palet,colorer,m);


		cur_xcor -= D_SPACE;
		cur_ycor -= D_SPACE;
		cur_level--;
		height_tracker = MINH;
		args_format.unset_print_tag(m->get_name());
		return base_box_width;
	}
	else {
		int this_box_width = 0;
		int this_box_height = 0;
		int this_box_deepest_level = cur_level;
		int temp_width_adder;
		int temp_height_tracker;
		int temp_xcur_offset = 0;
		int temp_ycur_offset = 0;
		std::vector<Box> box_draw_bounds;
		std::vector<VisualPort> this_box_VisualPortVector;
		int max_port = 0;
		int temp_port_max;		

		int name_width = (m->get_ast_type_name() + " " +  m->get_name()).length() * TEXT_CHAR;		
		int end_of_prev_word = name_width + cur_xcor;

		if(args_format.get_set_print_tag_true(m->get_name())) {
			if(args_format.is_drawing_ports()) {
			temp_ycur_offset = draw_space_channels(m,cur_ycor);
			}

			this_box_width += 2 * D_SPACE;
			this_box_width += (m->get_definition()->sub_modules_.size() - 1) * D_SPACE;
		}
		
		for(risc::ModuleInstanceVectorIter iter = m->get_definition()->sub_modules_.begin();
			iter != m->get_definition()->sub_modules_.end();
			iter++) {

				std::vector<VisualPort> temp_binded_ports;

				temp_width_adder = draw_recursor((*iter),this_box_deepest_level,temp_height_tracker,temp_binded_ports,cr,args_format);

				this_box_width += temp_width_adder;
				
				if(temp_height_tracker > this_box_height) {
					this_box_height = temp_height_tracker;
				}



				if(args_format.get_set_print_tag_true(m->get_name())) {

					if(args_format.is_drawing_ports()) {
						
						if(box_draw_bounds.empty() && (!m->get_definition()->sub_hier_channel_.empty() || !m->get_definition()->sub_prim_channel_.empty())) {
							draw_line(cr,0,0,0,0);
						}
						else {
							draw_line(NULL,0,0,0,0);
						}
						temp_port_max =  draw_compared_ports(cr,cur_xcor,cur_ycor - temp_ycur_offset,temp_binded_ports,VisualPortVector,m,(*iter),end_of_prev_word);
						if(max_port < temp_port_max) {
							max_port = temp_port_max;
						}				
						draw_channel_connect(cr, temp_binded_ports, m, (*iter),temp_ycur_offset, cur_ycor);
					}

				box_draw_bounds.push_back(Box(D_SPACE + cur_xcor,D_SPACE + cur_ycor,temp_width_adder,temp_height_tracker));
				

				if((*iter) != m->get_definition()->sub_modules_.back()) {
					temp_xcur_offset += D_SPACE + temp_width_adder;
					cur_xcor += D_SPACE + temp_width_adder;
				}
				}

		}
		if(args_format.get_set_print_tag_true(m->get_name())) {
		cur_xcor -= temp_xcur_offset;
		cur_ycor -= temp_ycur_offset;

		if(this_box_width < name_width) {
			this_box_width = name_width;
		}
		if(this_box_width < max_port - cur_xcor) {
			this_box_width = max_port - cur_xcor;
		}
	

		this_box_height = this_box_height + 2 * D_SPACE + temp_ycur_offset;
		draw_box(cr,cur_xcor,cur_ycor,this_box_width,this_box_height,m);
		if(!args_format.is_bw_only())color_module_dependent_box(cr,cur_xcor,cur_ycor,this_box_width,this_box_height,&box_draw_bounds,color_palet,colorer,m);
		
		if(args_format.is_drawing_ports()) draw_channels_wrapper(cr,m,cur_xcor,cur_ycor,this_box_width,temp_ycur_offset,VisualPortVector);

		if(this_box_deepest_level > param_deepest_level) {
			param_deepest_level = this_box_deepest_level;
		}

		cur_xcor -= D_SPACE;
		cur_ycor -= D_SPACE;
		cur_level--;
		args_format.unset_print_tag(m->get_name());
		}
		height_tracker = this_box_height;
		return this_box_width;
	}
}

void toggle_color(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_bw();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_quarter(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(0.25);

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_half(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(0.5);

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_full(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(1);

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_double(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->set_scale_factor(2);

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_increment(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->increment_scale_factor_by5percent();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void zoom_decrement(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->decrement_scale_factor_by5percent();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_increment(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->increment_depth();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_decrement(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->decrement_depth();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void depth_toggle(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_depth();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void port_toggle(GtkWidget* widget, gpointer* data) {
	ArgumentFormatter * args_formatter = (ArgumentFormatter*) data[2];

	args_formatter->toggle_ports();

	gtk_widget_hide((GtkWidget*) data[3]);
	gtk_widget_show_all((GtkWidget*) data[3]);
}

void ExposedEvent(GtkWidget *widget,GdkEvent *event, gpointer* data) {

		risc::Design* d = (risc::Design*) data[0]; 
		int deepest_level = 1;
		int top_width = 0;
		int top_height;
		std::vector<VisualPort> top_ports;
		
		ArgumentFormatter* pnt_passer = (ArgumentFormatter*) data[2];
		ArgumentFormatter& args_format = (*pnt_passer);
	

		cairo_t *cr = gdk_cairo_create(widget->window);	
	
		cairo_scale(cr,args_format.get_scale_factor(),args_format.get_scale_factor());
		cairo_set_source_rgb(cr,0.1,0.1,0.1);
		cairo_select_font_face(cr,"Purisa",CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
		cairo_set_font_size(cr,10);
		
		for(risc::InstanceVectorIter iter = d->top_modules_.begin();
			iter != d->top_modules_.end();
			iter++ ) {
			top_width += draw_recursor((*iter),deepest_level,top_height,top_ports,cr,args_format);
		}

		global_unscaled_top_width = top_width;
		global_unscaled_top_height = top_height;
	
		gtk_widget_set_size_request(widget,top_width * args_format.get_scale_factor() + 2 * D_SPACE,top_height * args_format.get_scale_factor() + 2 * D_SPACE);
	//gtk_widget_set_size_request((GtkWidget*)data[1],top_width + 4 * D_SPACE,top_height + 4 * D_SPACE);
}

void draw_page(GtkPrintOperation* operation, GtkPrintContext* context, gint pgNr,gpointer* data) {
	risc::Design* d = (risc::Design*) data[0]; 	

	int deepest_level = 1;
	int top_width = 0;
	int top_height;
	g_print("draw_page function called\n");
	std::vector<VisualPort> top_ports;

	ArgumentFormatter* pnt_passer = (ArgumentFormatter*) data[2];
	ArgumentFormatter& args_format = (*pnt_passer);
	
	cairo_t* cr = gtk_print_context_get_cairo_context(context);
	
	cairo_scale(cr,args_format.get_scale_factor(),args_format.get_scale_factor());

	cairo_set_source_rgb(cr,0.1,0.1,0.1);
	cairo_select_font_face(cr,"Purisa",CAIRO_FONT_SLANT_NORMAL,CAIRO_FONT_WEIGHT_BOLD);
	cairo_set_font_size(cr,10);

	for(risc::InstanceVectorIter iter = d->top_modules_.begin();
		iter != d->top_modules_.end();
		iter++ ) {
		top_width += draw_recursor((*iter),deepest_level,top_height,top_ports,cr,args_format);
	}

}

void do_print(GtkWidget* widget, gpointer* data) {
	std::cout << "In do_print!\n";

	GtkPrintOperation* print;
	GtkPrintSettings* settings;
	GtkPaperSize* paper_size;

	ArgumentFormatter* args_formatter = (ArgumentFormatter*) data[2];

	print = gtk_print_operation_new();
	gtk_print_operation_set_unit(print,GTK_UNIT_POINTS);
	gtk_print_operation_set_n_pages(print,1);
	gtk_print_operation_set_use_full_page(print,TRUE);
	gtk_print_operation_set_export_filename(print,"visualization.pdf");

	g_signal_connect(print,"draw_page",G_CALLBACK(draw_page),data);	

	paper_size = gtk_paper_size_new_custom("Visualization","display_Visualiztion",
						global_unscaled_top_width * args_formatter->get_scale_factor() + 2 * D_SPACE,global_unscaled_top_height * args_formatter->get_scale_factor() + 2 * D_SPACE,GTK_UNIT_POINTS);
	settings = gtk_print_settings_new();
	gtk_print_settings_set_paper_size(settings,paper_size);
	gtk_print_operation_set_print_settings(print,settings);



	gtk_print_operation_run(print,GTK_PRINT_OPERATION_ACTION_EXPORT,NULL,NULL);

}

int main(int argc, char* argv[]) {
	GtkWidget *window;
	GtkWidget *scrolled_window;
	GtkWidget * DrawArea;

	GtkWidget *menubar;
	GtkWidget *viewmenu;
	GtkWidget *view;
	GtkWidget *tog_stat;
	GtkWidget *zoomMenu;
	GtkWidget *zoom;
	GtkWidget *halfzoom;
	GtkWidget *quarterzoom;
	GtkWidget *fullzoom;
	GtkWidget *doublezoom;
	GtkWidget *sep;
	GtkWidget *incrementzoom;
	GtkWidget *decrementzoom;
	GtkWidget *depthmenu;
	GtkWidget *depth;
	GtkWidget *depthadd;
	GtkWidget *depthdecrease;
	GtkWidget *depthtoggle;	
	GtkWidget *toggleport;

	GtkWidget *file;
	GtkWidget *filemenu;
	GtkWidget *quit;
	GtkWidget *exportjpg;
	GtkWidget *print_btn;	

	void* pnt_buffer[4];

	ArgumentFormatter args_formatter(argc,argv);

	std::string LWIncludePath;
	LWIncludePath = "-I";
	LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
	LWIncludePath += "/include";
	
	// add -I$SYSTEMC_LW_HOME/include
	risc::tools::AddCommandlineOption(argc, argv, LWIncludePath);



	std::cout << std::endl
            	<< "Processing File..."
            	<< std::endl;

	risc::Design d(argc,argv);
	pnt_buffer[0] = &d;

	std::cout << std::endl
		<< "Finished Processing File."
		<< std::endl;

	gtk_init(&argc, &argv);

	pnt_buffer[2] = &args_formatter;

	window = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(window),"Module Visualization");
	gtk_window_set_default_size(GTK_WINDOW(window),300,300);
	gtk_window_set_resizable(GTK_WINDOW(window),TRUE);
	gtk_window_set_type_hint(GTK_WINDOW(window),GDK_WINDOW_TYPE_HINT_NORMAL);	
	pnt_buffer[1] = window;

	menubar = gtk_menu_bar_new();
	viewmenu = gtk_menu_new();
	filemenu = gtk_menu_new();
	zoomMenu = gtk_menu_new();
	depthmenu = gtk_menu_new();

	file = gtk_menu_item_new_with_label("File");
	print_btn = gtk_menu_item_new_with_label("Print");
	quit = gtk_menu_item_new_with_label("Quit");

	view = gtk_menu_item_new_with_label("View");
	quarterzoom = gtk_menu_item_new_with_label("25%");
	halfzoom = gtk_menu_item_new_with_label("50%");
	fullzoom = gtk_menu_item_new_with_label("100%");
	doublezoom = gtk_menu_item_new_with_label("200%");
	zoom = gtk_menu_item_new_with_label("Zoom");
	sep = gtk_separator_menu_item_new();
	incrementzoom = gtk_menu_item_new_with_label("Increment by 5%");
	decrementzoom = gtk_menu_item_new_with_label("Decrement by 5%");
	depth = gtk_menu_item_new_with_label("Depth");
	depthadd = gtk_menu_item_new_with_label("Increment Depth Level");
	depthdecrease = gtk_menu_item_new_with_label("Decrement Depth Level");
	depthtoggle = gtk_check_menu_item_new_with_label("Depth Variation Enabled");

	tog_stat = gtk_check_menu_item_new_with_label("No Color");
	if(args_formatter.is_bw_only()) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tog_stat), TRUE);
	}
	else {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(tog_stat), FALSE);
	}

	toggleport = gtk_check_menu_item_new_with_label("Draw Channels and Ports");
	if(args_formatter.is_drawing_ports()) {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(toggleport),TRUE);
	}
	else {
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(toggleport),FALSE);
	}

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(file),filemenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(filemenu),print_btn);
	gtk_menu_shell_append(GTK_MENU_SHELL(filemenu),quit);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), file);

	gtk_menu_item_set_submenu(GTK_MENU_ITEM(view),viewmenu);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(zoom),zoomMenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),quarterzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),halfzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),fullzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),doublezoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),sep);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),incrementzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(zoomMenu),decrementzoom);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),zoom);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(depth),depthmenu);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthadd);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthdecrease);
	gtk_menu_shell_append(GTK_MENU_SHELL(depthmenu),depthtoggle);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),depth);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu), tog_stat);
	gtk_menu_shell_append(GTK_MENU_SHELL(viewmenu),toggleport);
	gtk_menu_shell_append(GTK_MENU_SHELL(menubar), view);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox), menubar, FALSE, FALSE, 0);


	scrolled_window = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),GTK_POLICY_AUTOMATIC,GTK_POLICY_ALWAYS);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(window)->vbox),scrolled_window,TRUE,TRUE,0);

	DrawArea = gtk_drawing_area_new();
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window),DrawArea);
	pnt_buffer[3] = DrawArea;

	g_signal_connect(G_OBJECT(DrawArea),"expose_event",G_CALLBACK(ExposedEvent),pnt_buffer);

	g_signal_connect(G_OBJECT(window),"destroy",G_CALLBACK(gtk_main_quit),NULL);

	g_signal_connect(G_OBJECT(quit),"activate",G_CALLBACK(gtk_main_quit),NULL);

	g_signal_connect(G_OBJECT(tog_stat), "activate", G_CALLBACK(toggle_color),pnt_buffer);

	g_signal_connect(G_OBJECT(print_btn), "activate", G_CALLBACK(do_print),pnt_buffer);

	g_signal_connect(G_OBJECT(quarterzoom),"activate",G_CALLBACK(zoom_quarter),pnt_buffer);

	g_signal_connect(G_OBJECT(halfzoom),"activate",G_CALLBACK(zoom_half),pnt_buffer);

	g_signal_connect(G_OBJECT(fullzoom),"activate",G_CALLBACK(zoom_full),pnt_buffer);

	g_signal_connect(G_OBJECT(doublezoom),"activate",G_CALLBACK(zoom_double),pnt_buffer);

	g_signal_connect(G_OBJECT(incrementzoom),"activate",G_CALLBACK(zoom_increment),pnt_buffer);
	
	g_signal_connect(G_OBJECT(decrementzoom),"activate",G_CALLBACK(zoom_decrement),pnt_buffer);

	g_signal_connect(G_OBJECT(depthadd),"activate",G_CALLBACK(depth_increment),pnt_buffer);

	g_signal_connect(G_OBJECT(depthdecrease),"activate",G_CALLBACK(depth_decrement),pnt_buffer);

	g_signal_connect(G_OBJECT(depthtoggle),"activate",G_CALLBACK(depth_toggle),pnt_buffer);

	g_signal_connect(G_OBJECT(toggleport),"activate",G_CALLBACK(port_toggle),pnt_buffer);

	gtk_widget_show_all(window);

	gtk_main();	
}
