#ifndef DESGIN_H_INCLUDED_
#define DESGIN_H_INCLUDED_

#include "rose.h"
#include "port_mapping_cache.h"

#include <vector>

#define PSG_MODE risc::get_design()->psg_mode

namespace risc {

class Thread;
class Event;
class Function;
class HierarchicalChannel;
class HierarchicalChannelInstance;
class Instance;
class InstanceTree;
class Interface;
class Module;
class PathInstanceMapper;
class PrimitiveChannel;
class PrimitiveChannelInstance;
class Variable;

typedef std::vector<Event*> EventVector;
typedef std::vector<Function*> FunctionVector;
typedef std::vector<HierarchicalChannel*> HierarchicalChannelVector;
typedef std::vector<HierarchicalChannelInstance*> HierarchicalChannelInstanceVector;
typedef std::vector<Instance*> InstanceVector;
typedef std::vector<Interface*> InterfaceVector;
typedef std::vector<Module*> ModuleVector;
typedef std::vector<PrimitiveChannel*> PrimitiveChannelVector;
typedef std::vector<PrimitiveChannelInstance*> PrimitiveChannelInstanceVector;
typedef std::vector<Variable*> VariableVector;

/**
 * This class represents the internal representation. The IR offers different
 * interfaces to start search in IR.
 */
class Design {

public:

  /**
   * \brief default constructor
   * \param argc  expects argc from main function
   * \param argv  expects argv from main function
   */
  explicit Design(int argc, char* argv[], bool is_psg_mode = false);
  ~Design();

  void initialize_port_mapping();
  void initialize_interfaces();
  void initialize_definitions();
  void initialize_top_modules();
  void initialize_global_functions();
  void initialize_global_variables();
  void initialize_path_instance_mapper();

  /*
   * Functions to print design properties
   */
  void print_all_class_definitions();
  void print_all_global_functions();
  void print_design();

  /*
   * Interface for the instance tree
   */
  InstanceTree get_root();

  /*
   * Containers for the different design elements
   */
  ModuleVector module_definitions_;
  InstanceVector top_modules_;
  PrimitiveChannelVector primitive_channel_;
  HierarchicalChannelVector hierarchical_channel_;
  InterfaceVector interfaces_;

  /*
   * global and sc_main related design stuff
   */
  VariableVector global_variables_;
  FunctionVector global_functions_;

  /*
   * Functions to search in design
   */
  Module *find_module_definition(SgClassDefinition* cd);
  PrimitiveChannel* find_prim_channel_definition(SgClassDefinition* cd);
  HierarchicalChannel* find_channel_definition(SgClassDefinition* cd);
  Interface *find_interface(SgClassDefinition* cd);

  /**
   * \brief Find a port by it's ast variable definition
   */
  Variable* find_global_variable(SgVariableDefinition* var_def);

  /*
   * Rose specific elements
   */
  SgProject *project_;

  /*
   * misc
   */
  int old_argc_;
  char **old_argv_;

  std::vector<std::string> input_files_;

  PathInstanceMapper *path_instance_mapper_;

  PortMappingCache port_mapping_cache_;

  /**
   * \fn    reset_instance_counter
   * \brief This function resets the instance counter of the module and
   *        hierarchical channel instances to zero.
   */
  void reset_instance_counter();

  /**
   * \brief This variable contains the filename of the input file for the
   *        dynamic analysis
   */
  std::string dynamic_analysis_filename_;

  /**
   * \brief This variable contains the name which will be used for the dir
   *        file during the presimulation
   */
  std::string instrumentor_output_filename_;

  /**
   * \fn    function perform_static_analysis
   * \brief The functions return if a static analysis will be performed
   */
  bool perform_static_analysis();

  /**
   * \fn    add_module_to_definition_list
   * \brief Only for internal use!
   *        This function triggers the analysis of a module.
   *        The new analyzed module will be put into the module_definitions_
   *        container
   */
  void add_module_to_definition_list(SgClassDefinition *class_def);

  /**
   * \fn    add_hierarchical_channel_to_definition_list
   * \brief Only for internal use!
   *        This function triggers the analysis of a module.
   *        The new analyzed module will be put into the
   *        hierarchical_channel_ container
   */
  void add_hierarchical_channel_to_defintion_list(SgClassDefinition *class_def);

  /**
   * \fn    file_name
   * \brief This function returns the input file name of the design
   */
  std::string get_design_file_name();

  bool is_analyzable();
  
  bool psg_mode;
  int psg_count;
  std::vector<int> psg_seg_count;
  std::vector<Thread*> loaded_threads; 
  std::vector<Thread*> simulation_threads;

private:

  /**
   * \brief Disabled copy constructor
   */
  Design(const Design &d);
};

class GlobalDesign {
public:
  static Design *design_static;
};
Design* get_design();
SgProject* get_project();

}; // end namespace risc

#endif /* DESGIN_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
