#ifndef DYNAMIC_PATH_INSTANCE_MAPPER_H_INCLUDED_
#define DYNAMIC_PATH_INSTANCE_MAPPER_H_INCLUDED_

#include <vector>

#include "instance_tree.h"
#include "path_instance_mapper.h"

class DynamicModule;

namespace risc {

/**
 * \class PathInstanceMapper
 * \brief This class can be used as a lookup table between an instance id of
 *        a module or hierarchical channel and the instance tree.
 *
 *        There is a significant difference between the PathInstanceMapper and
 *        the DynamicPathInstanceMapper. The DynamicPathInstanceMapper will be
 *        generated from the pre simulation of evaluation.
 */
class DynamicPathInstanceMapper: public PathInstanceMapper {

public:

  /**
   * \fn       PathInstanceMapper
   * \brief    This is the default constructor and need a design and a segment
   *           graph. The lookup table will be automatically created.
   */
  DynamicPathInstanceMapper(std::string filename);

  void traverse_tree(InstanceTree root, DynamicModule *module);

  /**
   * \fn       determine_lookup_table
   * \brief    This function traverses the design and pushes all modules and
   *           hierarchical channel instances into the path_to_id_mapping_
   *           lookup table. The traversing is dependent on the declaration
   *           order of the member variables. Also, it follows depth first
   *           search.
   *           CAUTION: The function Design::reset_instance_counter() must be
   *                    used first! Otherwise wrong intance id's a generated!
   */
  void determine_lookup_table();

  /**
   * \brief This is the input file for the dynamic analysis
   */
  std::string filename_;

  /**
   * \brief This variables stores the root node of the dynamic analysis
   */
  DynamicModule *dynamic_root_;

  /**
   * \brief This list stores the global variables of the dynamic analysis
   */
  std::list<std::pair<std::string, int*> > global_variable_addresses_;

  /**
   * \brief This functions parses the dynamic analysis file into a tree
   *        representation
   */
  void parse_dynamic_instance_tree();

  /*
   * \fn       function get_dynamic_tree_root
   * \brief    This function returns the root node of the dynamic analysis
   */
  virtual DynamicModule* get_dynamic_tree_root();

  /**
   * \fn       perform_dynamic_analysis
   * \brief    This functions returns true if the path instance mapper should
   *           perform a dynamic analysis
   */
  virtual bool perform_dynamic_analysis();

  /**
   * \fn       get_global_variable_address()
   * \brief    This functions returns the address of a global variable.
   */
  virtual int* get_global_variable_address(std::string variable_name);

  /**
   * \fn       get_global_variable_name()
   * \brief    This functions returns the name of a global variable.
   */
  virtual std::string get_global_variable_name(int *variable_address);

  /**
   * \fn       translate_type_name
   * \brief    This function is only for internal use.
   *           Rose cannot generate type names for template classes.
   *           The output is very cryptic. So this function translates a clean
   *           gnu class name into a rose type name.
   */
  std::string translate_type_name(std::string gnu_type_name);
};

} // end of namespace risc

#endif /* DYNAMIC_PATH_INSTANCE_MAPPER_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
