#include "static_path_instance_mapper.h"

#include "design.h"
#include "definition.h"
#include "hierarchical_channel.h"
#include "hierarchical_channel_instance.h"
#include "module.h"
#include "module_instance.h"

#include "../segment_graph/segment_graph.h"

risc::StaticPathInstanceMapper::StaticPathInstanceMapper(InstanceTree root)
{
  determine_lookup_table(root);
}

void
risc::StaticPathInstanceMapper::determine_lookup_table(InstanceTree root)
{
  Instance *instance = dynamic_cast<risc::Instance*>(root.get_object());

  if(instance) {

    // We have to avoid primitive channels because they don't contain a
    // sub hierarchical components
    if(instance->type_ == Definition::PrimitiveChannelInstanceType) {
      return;
    }

    Class *class_definition = instance->get_definition();

    // We are iterating over all declarations in the channel/module.
    // We cannot use the internal infrastructure because we need to do it in
    // the depth first search order.
    for(SgDeclarationStatementPtrList::iterator
        iter  = class_definition->get_ast_node()->get_members().begin();
        iter != class_definition->get_ast_node()->get_members().end();
        iter++) {

      // we have to consider all variable declarations
      SgVariableDeclaration *var_decl = isSgVariableDeclaration(*iter);
      if(var_decl) {

        Instance *sub_instance
          = instance->get_definition()->find_instance(
                                            var_decl->get_definition());

        if(sub_instance) {
          // we are going recursively through the design

          risc::InstanceTree::Element sub_path;

          sub_path.first = instance->get_definition();
          sub_path.second = sub_instance;

          root.push_back(sub_path);
          determine_lookup_table(root);
          root.pop_back();
        }
      }
    }
  }

  risc::HierarchicalChannelInstance *hier_channel_instance
    = dynamic_cast<risc::HierarchicalChannelInstance*>(root.get_object());

  if(hier_channel_instance) {

    path_to_id_mapping_.push_back(
        PathWithID(root,
            hier_channel_instance->get_definition()->instance_counter_));

    hier_channel_instance->get_definition()->instance_counter_++;

    if(max_instances_ <
        hier_channel_instance->get_definition()->instance_counter_) {
      max_instances_
        = hier_channel_instance->get_definition()->instance_counter_;
    }
    return;
  }

  risc::ModuleInstance *module_instance
    = dynamic_cast<risc::ModuleInstance*>(root.get_object());

  if(module_instance) {

    path_to_id_mapping_.push_back(
        PathWithID(root, module_instance->get_definition()->instance_counter_));

    module_instance->get_definition()->instance_counter_++;

    if(max_instances_ < module_instance->get_definition()->instance_counter_) {
      max_instances_ = module_instance->get_definition()->instance_counter_;
    }
    return;
  }
}


int
risc::StaticPathInstanceMapper::get_id(InstanceTree tree)
{
  // The path instance mapper stores only path which end with
  // a module or channel.
  // So, we have shorten the path if there is any other element at the end.
  while((tree.path_.back().second->type_
        != risc::Definition::ChannelInstanceType) &&
      (tree.path_.back().second->type_
        != risc::Definition::HierarchicalChannelInstanceType) &&
      (tree.path_.back().second->type_
        != risc::Definition::ModuleInstanceType) ) {

    tree.pop_back();
  }

  for(std::vector<PathWithID>::iterator
      iter  = path_to_id_mapping_.begin();
      iter != path_to_id_mapping_.end();
      iter++) {

    if((*iter).first == tree)
      return (*iter).second;
  }

  assert("Not matching instance tree" && 0);
}

risc::InstanceTree
risc::StaticPathInstanceMapper::get_instance_tree(
    int id,
    std::list<InstanceTree> potential_instances)
{
  for(std::list<InstanceTree>::iterator
      potential_iter  = potential_instances.begin();
      potential_iter != potential_instances.end();
      potential_iter++) {

    InstanceTree temp = (*potential_iter).get_parent();

    for(std::vector<PathWithID>::iterator
        mapping_iter  = path_to_id_mapping_.begin();
        mapping_iter != path_to_id_mapping_.end();
        mapping_iter++) {

      if((*mapping_iter).second == id &&
          (*mapping_iter).first == temp) {

        return *potential_iter;
      }
    }
  }

  // in this case there was no matching
  return InstanceTree();
}

bool
risc::StaticPathInstanceMapper::perform_dynamic_analysis()
{
  return false;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
