#include "graph_helper.h"

#include "port_call_path.h"

std::list<int>
risc::sg::get_all_reachable_segments(SegmentGraph &segment_graph,
    int starting_id)
{
  std::pair<VertexIterator, VertexIterator> vp;

  VertexIterator starting_vertex;

  // Here we are searching for the for the segment with the id 'starting_id'
  for(vp = boost::vertices(segment_graph.graph_);
      vp.first != vp.second; ++vp.first) {

    if(segment_graph.graph_[*(vp.first)].id_ == starting_id) {
      starting_vertex = vp.first;
      break;
    }
  }

  // Reseting all segments to the status untouched
  segment_graph.set_all_segments_to_untouched();

  risc::sg::SegmentGraph::SegmentSet untouched;
  untouched.insert(*starting_vertex);

  std::list<int> reachable_segment_ids;

  // We are iterating until there are no more untouched segments
  while(!untouched.empty()) {

    VertexDescriptor vertex = *untouched.begin();

    // Marking the current segment as touched
    segment_graph.graph_[vertex].helper_for_graph_algorithms_
      = Segment::Touched;

    // Adding the id to the return list of reachable segment ids
    reachable_segment_ids.push_back(segment_graph.graph_[vertex].id_);

    Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
    boost::tie(out_edge_begin, out_edge_end)
      = boost::out_edges(vertex, segment_graph.graph_);

    // Iterating over all outgoing edges to discover new segments
    for(out_edge_iter  = out_edge_begin;
        out_edge_iter != out_edge_end;
        out_edge_iter++) {

      if(segment_graph.graph_[*out_edge_iter].event_notification_edge_ ||
          segment_graph.graph_[*out_edge_iter].hierarchical_communication_) {
        continue;
      }

      VertexDescriptor target
        = boost::target(*out_edge_iter, segment_graph.graph_);

      // If the segment is untouched, we have to add it to the set of untouched
      // segments. We will evaluate this segment in one of the next iterations.
      if(segment_graph.graph_[target].helper_for_graph_algorithms_
          == Segment::Untouched) {

        untouched.insert(target);
      }
    }

    // Removing the touched segment from the untouched segment set
    untouched.erase(vertex);
  }

  return reachable_segment_ids;
}



std::vector<risc::sg::VertexDescriptor>
risc::sg::get_all_reachable_segments(
    SegmentGraph &segment_graph,
    risc::sg::VertexDescriptor start_vertex)
{

  // Reseting all segments to the status untouched
  segment_graph.set_all_segments_to_untouched();

  risc::sg::SegmentGraph::SegmentSet untouched;
  untouched.insert(start_vertex);

  std::vector<risc::sg::VertexDescriptor> reachable_segments;

  // We are iterating until there are no more untouched segments
  while(!untouched.empty()) {

    VertexDescriptor vertex = *untouched.begin();

    // Removing the touched segment from the untouched segment set
    untouched.erase(vertex);

    // Marking the current segment as touched
    segment_graph.graph_[vertex].helper_for_graph_algorithms_
      = Segment::Touched;

    // Adding the id to the return list of reachable segment ids
    reachable_segments.push_back(vertex);


    Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
    boost::tie(out_edge_begin, out_edge_end)
      = boost::out_edges(vertex, segment_graph.graph_);

    // Iterating over all outgoing edges to discover new segments
    for(out_edge_iter  = out_edge_begin;
        out_edge_iter != out_edge_end;
        out_edge_iter++) {


      VertexDescriptor target
        = boost::target(*out_edge_iter, segment_graph.graph_);

      // If the segment is untouched, we have to add it to the set of untouched
      // segments. We will evaluate this segment in one of the next iterations.
      if(segment_graph.graph_[target].helper_for_graph_algorithms_
          == Segment::Untouched) {

        untouched.insert(target);
      }
    }

  }

  return reachable_segments;
}

std::pair<risc::sg::VertexDescriptor, risc::sg::SegmentGraph::SegmentSet>
risc::sg::clone_graph(
    risc::sg::VertexDescriptor start_vertex,
    risc::sg::SegmentGraph::SegmentSet leaving_vertices,
    risc::sg::SegmentGraph &sg,
    bool channel_segments)
{
  std::vector<VertexDescriptor> reachable
    = get_all_reachable_segments(sg, start_vertex);
  std::vector<VertexDescriptor> clone;

  // here we duplicate all segments
  for(std::vector<VertexDescriptor>::iterator
      iter  = reachable.begin();
      iter != reachable.end();
      iter++) {

    Segment s(sg.graph_[*iter]);
    s.id_ = risc::sg::Segment::id_counter;
    s.node_id_ = risc::sg::Segment::node_counter;
    s.is_channel_segment_ = channel_segments;
    risc::sg::Segment::id_counter++;
    risc::sg::Segment::node_counter++;
    VertexDescriptor segment_vertex = boost::add_vertex(s, sg.graph_);
    clone.push_back(segment_vertex);
  }

  // here, we add all the edges
  for(int i = 0; i < reachable.size(); i++) {

    Graph::out_edge_iterator out_edge_begin, out_edge_end, out_edge_iter;
    boost::tie(out_edge_begin, out_edge_end) =
      boost::out_edges(reachable[i], sg.graph_);

    // we add for every node the outgoing edges
    // it is not necessary to add the ingoing edges
    // because they are also outgoing edges
    for(out_edge_iter  = out_edge_begin;
        out_edge_iter != out_edge_end;
        out_edge_iter++) {

      VertexDescriptor target = boost::target(*out_edge_iter, sg.graph_);

      int target_index = -1;
      for(int j=0; j < reachable.size(); j++) {
        if(sg.graph_[reachable.at(j)].id_ == sg.graph_[target].id_) {
          target_index = j;
          break;
        }
      }

      boost::add_edge(clone.at(i), clone.at(target_index), sg.graph_);
    }
  }

  std::pair<VertexDescriptor, risc::sg::SegmentGraph::SegmentSet> result;

  // setting entering segment
  int index = -1;
  for(int j=0; j < reachable.size(); j++) {
    if(sg.graph_[reachable.at(j)].id_ == sg.graph_[start_vertex].id_) {
      index = j;
      break;
    }
  }
  result.first = clone.at(index);

  // setting leaving segments
  for(risc::sg::SegmentGraph::SegmentSet::iterator
      iter  = leaving_vertices.begin();
      iter != leaving_vertices.end();
      iter++) {

    int index = -1;
    for(int j=0; j < reachable.size(); j++) {
      if(sg.graph_[reachable.at(j)].id_ == sg.graph_[*iter].id_) {
        index = j;
        break;
      }
    }

    result.second.insert(clone.at(index));
  }

  return result;
}

void
risc::sg::add_pcp_to_graph(
    risc::sg::VertexDescriptor start_vertex,
    risc::sg::SegmentGraph &sg,
    SgFunctionCallExp *func_call_exp)
{
  std::vector<VertexDescriptor> reachable
    = get_all_reachable_segments(sg, start_vertex);

  // here we duplicate all segments
  for(std::vector<VertexDescriptor>::iterator
      seg_iter  = reachable.begin();
      seg_iter != reachable.end();
      seg_iter++) {

    // Adding the port call path to the boundary statement
    sg.graph_[*seg_iter].pcp_.path_.push_back(func_call_exp);

    for(std::list<NodeWithPath>::iterator
        expr_iter  = sg.graph_[*seg_iter].expressions_.begin();
        expr_iter != sg.graph_[*seg_iter].expressions_.end();
        expr_iter++) {

      expr_iter->pcp_.path_.push_back(func_call_exp);
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
