#ifndef PREDICTION_EVENT_NOTIFICATION_TABLE_H_INCLUDED_
#define PREDICTION_EVENT_NOTIFICATION_TABLE_H_INCLUDED_

#include<string>
#include<utility>
#include<vector>

#include "event_conflict_table.h"
#include "time.h"

#include "segment_graph.h" // needed for VertexDescriptor
#include "mapped_variable.h" // needed for MappedVariable

namespace risc {

namespace sg {

class PredictionEventNotificationTable {

public:

  /**
   * \fn    constructor
   * \brief Creates the combined data conflict table.
   *        The maximum number of steps will be calculated automatically
   */
  PredictionEventNotificationTable(
      risc::sg::EventConflictTable &ect,
      PathInstanceMapper *path_instance_mapper);

  /**
   * \fn       get_max_instances
   * \brief    This functions returns how many instances of a given segment
   *           exist
   */
  int get_max_instances(int segment_id);

  /**
   * \fn       index_to_segment_and_instance_id
   * \brief    This function translates a given index from the conflict table
   *           into the corresponding segment id and instance id. The first
   *           value is the segment id and the second value is the instance id.
   */
  std::pair<int, int> index_to_segment_and_instance_id(int index);

  /**
   * \fn       segment_and_instance_id_to_index
   * \brief    This function translates a given segment id and instance id
   *           to a index of the conflict table
   */
  int segment_and_instance_id_to_index(int segment_id, int instance_id);

  /**
   * \fn    print_notification_table
   * \brief This function prints the prediction event notification table
   *        in the terminal
   */
  void print_notification_table();

  /**
   * \fn    print_notification_table
   * \brief This function prints the prediction event notification table
   *        in the file filename
   */
  void print_notification_table(std::string filename);

  /**
   * \fn    print_instance_id_to_index_lookup_table
   * \brief This function prints the table instance_id_to_index_lookup_ table
   *        in the terminal
   */
  void print_instance_id_to_index_lookup_table();

  /**
   * \fn    print_instance_id_to_index_lookup_table
   * \brief This function prints the table instance_id_to_index_lookup_ table
   *        in the file filename
   */
  void print_instance_id_to_index_lookup_table(std::string filename);


  PathInstanceMapper *path_instance_mapper_;

  /**
   * \brief This lookup table translates segment ids into vertex descriptors.
   *        At position i is the segment with the id of
   *        instance_id_to_index_lookup_[0][i]
   */
  VertexDescriptor *vertex_lookup_;

  /**
   * \brief This variable represents the max number of module instances in the
   *        design. Also, this is the number of rows in the
   *        instance_id_to_index_lookup_ table.
   */
  int max_instances_;

  /**
   * \brief This variable defines how many segments exist in the segment
   *        graph. Also, this is the number of columns in the
   *        instance_id_to_index_lookup_ table.
   */
  int number_of_segments_;

  /**
   * \brief This variable defines the size of the conflict table.
   *        The size is quadratic.
   */
  int size_of_conflict_table_;

  /**
   * \brief This table should be used a lookup table:
   *        (segment id, instance id) -> index of conflict table
   */
  int **instance_id_to_index_lookup_;

  /**
   * \brief The prediction event notification table
   */
  Time **combined_table_;

  risc::sg::EventConflictTable &ect_;

private:

  struct Node {
    VertexDescriptor vd;
    Time t;
    int inst;

    Node(VertexDescriptor vd, Time t, int inst):
      vd(vd), t(t), inst(inst)
    { }

    Node(const Node &ohs):
      vd(ohs.vd), t(ohs.t), inst(ohs.inst)
    { }
  };

  typedef std::pair<VertexDescriptor, Time> SegmentTimeTuple;

  /**
   * \fn
   * \brief
   */
  void bfs(VertexDescriptor start, int instance);

  /**
   * \fn    compute_combined_table
   * \brief This function computes the combined table for 'prediction_steps_'
   *        prediction steps
   */
  void compute_combined_table();

};

} // end of namespace sg

} // end of namespace risc

#endif /* PREDICTION_EVENT_NOTIFICATION_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
